//
//  NisusRangeUtil.h
//  NisusCore
//
//  Created by Martin on 2013.10.03.
//  Copyright (c) 2013 Nisus. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <NisusCore/NisusAssertions.h>

// useful ranges
#define NisusRange_ZeroRange (NSMakeRange(0,0))
#define NisusRange_NotFoundRange (NSMakeRange(NSNotFound,0))



// Getting first/last locations in a range, which must not be of zero length.
#define NisusRange_FirstLocationIn( rangeExpr )				\
	({														\
		NSRange nisusRange = (rangeExpr);					\
		NISUS_ASSERT( 0 != nisusRange.length, "ranges" );	\
		NSUInteger nisusResult = nisusRange.location;		\
		nisusResult;										\
	})

#define NisusRange_LastLocationIn( rangeExpr )					\
	({															\
		NSRange nisusRange = (rangeExpr);						\
		NISUS_ASSERT( 0 != nisusRange.length, "ranges" );		\
		NSUInteger nisusResult = NSMaxRange(nisusRange) - 1;	\
		nisusResult;											\
	})



/************************************
 * Making Ranges
 ************************************/

// Creates a range that includes "start" and "end", inclusive.
// PRE-CON: start must be before (or equal to) end.
#define NisusRange_MakeInclusive( startExpr, endExpr )								\
	({																				\
		NSUInteger nisusStart = (startExpr);										\
		NSUInteger nisusEnd = (endExpr);											\
		NISUS_ASSERT( nisusStart <= nisusEnd, "ranges" );							\
		NSRange nisusResult = NSMakeRange( nisusStart, nisusEnd - nisusStart + 1 );	\
		nisusResult;																\
	})

// Creates a range that includes "start" and continues up to, but does not include, "bound".
#define NisusRange_AtWithBound( atExpr, boundExpr )								\
	({																			\
		NSUInteger nisusAt = (atExpr);											\
		NSUInteger nisusBound = (boundExpr);									\
		NISUS_ASSERT( nisusBound >= nisusAt, "ranges" );						\
		NSRange nisusResult = NSMakeRange( nisusAt, (nisusBound - nisusAt) );	\
		nisusResult;															\
	})

// Returns the range that includes everything after the bound of the first range, up to but not including the bound of the second range.
// If the first range's bound is greater than (or equal to) the bound of the second range, this returns the zero-length range after the first range.
static NISUS_INLINE NSRange NisusRange_BetweenBounds( NSRange firstRange, NSRange secondRange )
{
	NSUInteger start = NSMaxRange(firstRange);
	NSUInteger bound = NSMaxRange(secondRange);
	NSUInteger len = (start >= bound) ? 0 : (bound - start);
	return NSMakeRange( start, len );
}

// Returns the zero length range with the given location
static NISUS_INLINE NSRange NisusRange_ZeroRangeAt( NSUInteger loc )
{
	return NSMakeRange(loc,0);
}

// Returns the zero length range at the end of the given string or attributed string.
#define NisusRange_ZeroRangeAfterText(textObj) (NisusRange_ZeroRangeAt([(textObj) length]))

// Returns a range that fully encloses the given string or attributed string.
#define NisusRange_ForText(textObj) (NSMakeRange(0, [(textObj) length]))

// Returns zero length ranges that start at either "rng.location" or "NSMaxRange(rng)".
static NISUS_INLINE NSRange NisusRange_ZeroRangeBefore( NSRange rng )
{
	return NSMakeRange(rng.location, 0);
}

static NISUS_INLINE NSRange NisusRange_ZeroRangeAfter( NSRange rng )
{
	return NSMakeRange(NSMaxRange(rng), 0);
}



/************************************
 * Calculations
 ************************************/

// Alternative to NSIntersectionRange that handles zero length ranges.
extern NSRange NisusRange_IntersectionRange( NSRange aRange, NSRange bRange );

// Returns YES if the ranges intersect.
// If the first range is zero length, still returns YES if its location is contained in the second range.
static NISUS_INLINE BOOL NisusRange_IsIntersected( NSRange smallRange, NSRange wholeRange )
{
    if( 0 == smallRange.length ) {
        return (smallRange.location >= wholeRange.location) && (smallRange.location < NSMaxRange(wholeRange));
    }
    return (0 != NSIntersectionRange(smallRange, wholeRange).length);
}

// Tests if the first range completely contains the other
static NISUS_INLINE BOOL NisusRange_WholeContainsSub( NSRange wholeRng, NSRange subRng )
{
	if( 0 == subRng.length ) {
		if( 0 == wholeRng.length ) return (subRng.location == wholeRng.location);
		return (subRng.location >= wholeRng.location) && (subRng.location <= NSMaxRange(wholeRng));
	}
	return (subRng.length == NSIntersectionRange(wholeRng, subRng).length);
}

// Sorts ranges by starting location. If same location, then shorter ranges come first.
NSComparisonResult NisusRange_Compare( NSRange rng1, NSRange rng2 );
NSComparisonResult NisusRange_CompareByBound( NSRange rng1, NSRange rng2 );

// math operations
NSRange NisusRange_UnionContiguous( NSRange rng1, NSRange rng2 );
NSUInteger NisusRange_SubtractRange( NSRange wholeRange, NSRange subtractRange, NSRange* results );
NSArray* NisusRange_RangesBySubtractingRange( NSRange wholeRange, NSRange subtractRange );

// This function will return the union of the two ranges, but ignores an existing zero length range.
// If the first range is zero length, the second range is always returned as-is.
static NISUS_INLINE NSRange NisusRange_UnionIgnoringFirstEmpty( NSRange curRange, NSRange addRange )
{
	if( 0 == curRange.length ) {
		return addRange;
	}
	return NSUnionRange(curRange, addRange);
}

// This function only calculates the union if both ranges are non-zero length.
// Any zero length ranges are ignored. If both ranges are zero-lenght returns NisusRange_NotFoundRange.
static NISUS_INLINE NSRange NisusRange_UnionIgnoringEmpty( NSRange curRange, NSRange addRange )
{
	if( 0 == curRange.length ) {
		if( 0 == addRange.length ) {
			return NisusRange_NotFoundRange;
		}
		return addRange;
	}
	else if( 0 == addRange.length ) {
		return curRange;
	}
	return NSUnionRange( curRange, addRange );
}

// Adds an index to the range, no matter if the index comes before or after the input range. Ignores the input range if its location is NSNotFound.
static NISUS_INLINE NSRange NisusRange_UnionIndex( NSRange curRange, NSUInteger addIndex )
{
	if( NSNotFound == curRange.location ) {
		return NSMakeRange(addIndex,1);
	}
	
	if( addIndex > NSMaxRange(curRange) ) {
		curRange.length = (addIndex + 1) - curRange.location;
	}
	else if( addIndex < curRange.location ) {
		curRange = NisusRange_AtWithBound( addIndex, NSMaxRange(curRange) );
	}
	return curRange;
}

// Just like NisusRange_AddIndex but assumes the given index always comes after the input range.
static NISUS_INLINE NSRange NisusRange_UnionIndexAfter( NSRange curRange, NSUInteger addIndex )
{
	if( NSNotFound == curRange.location ) {
		return NSMakeRange(addIndex,1);
	}
	NISUS_ASSERT( addIndex >= curRange.location, "ranges" );
	curRange.length = (addIndex + 1) - curRange.location;
	return curRange;
}




/************************************
 Clipping and Extending Ranges
 ************************************/

static NISUS_INLINE BOOL NisusRange_IsInLimit( NSRange range, NSUInteger limit )
{
	NSUInteger bound = NSMaxRange(range);
	if( (bound > limit) || (bound < range.location) ) { // check for overflow too!
		return NO;
	}
	return YES;
}

// Ensures that the given range's bound falls at or before the given limit
static NISUS_INLINE NSRange NisusRange_ClipToLimit( NSRange range, NSUInteger limit )
{
	if( NisusRange_IsInLimit(range, limit) ) {
		return range;
	}
	NSRange fixed;
	fixed.location = MIN( range.location, limit );
	fixed.length = limit - fixed.location;
	return fixed;
}

// Does the same as NisusRange_ClipToLimit but will assert if the range required any clipping.
#define NisusRange_AssertToLimit( rangeExpr, limitExpr )						\
	({																			\
		NISUS_ASSERT( NisusRange_IsInLimit(rangeExpr,limitExpr), "ranges" );	\
		NSRange nisusResult = NisusRange_ClipToLimit( rangeExpr, limitExpr );	\
		nisusResult;															\
	})


// Returns a version of the given range with the first "clipCount" items removed. If the clipping exceeds the length of
// the range then a zero length range is returned that starts just after the original range.
static NISUS_INLINE NSRange NisusRange_ClipStart( NSRange rng, NSUInteger clipCount )
{
	if( clipCount > rng.length ) clipCount = rng.length;
	return NSMakeRange(rng.location + clipCount, rng.length - clipCount);
}

static NISUS_INLINE NSRange NisusRange_ClipEnd( NSRange rng, NSUInteger clipCount )
{
	if( clipCount > rng.length ) clipCount = rng.length;
	return NSMakeRange(rng.location, rng.length - clipCount);
}

// Adds the desired number of items to the start of the range, eg: length is increased and endpoint remains the same.
// If the range cannot accomodate the addition then increases by maximum amount allowed.
static NISUS_INLINE NSRange NisusRange_ExtendStart( NSRange rng, NSUInteger count )
{
	if( count > rng.location ) count = rng.location;
	rng.location -= count;
	rng.length += count;
	return rng;
}

// Same but extends the end of the range.
static NISUS_INLINE NSRange NisusRange_ExtendEnd( NSRange rng, NSUInteger count )
{
	rng.length += count;
	return rng;
}

static NISUS_INLINE NSRange NisusRange_ExtendEndWithLimit( NSRange aRange, NSUInteger count, NSUInteger limit )
{
	NSRange newRange = NSMakeRange( aRange.location, aRange.length + count );
	newRange = NisusRange_ClipToLimit( newRange, limit );
	return newRange;
}

// Expands the given range to include the provided location, which can be either before or after the range (or inside the range)
static NISUS_INLINE NSRange NisusRange_ExtendToLocation( NSRange rng, NSUInteger loc )
{
	if( loc < rng.location ) {
		rng.length += (rng.location - loc);
		rng.location = loc;
	}
	else if( loc >= NSMaxRange(rng) ) {
		rng.length = loc - rng.location + 1;
	}
	return rng;
}

// Expands the start and end of the range, but within the given limit.
static NISUS_INLINE NSRange NisusRange_Extend( NSRange aRange, NSUInteger extendStart, NSUInteger extendEnd, NSUInteger limit )
{
	NSRange newRange = NisusRange_ExtendStart( aRange, extendStart );
	newRange = NisusRange_ExtendEndWithLimit( newRange, extendEnd, limit );
	return newRange;
}



// Calculate ranges adjusted for edits.
NSRange NisusRange_AdjustedPerfectlyForEdit( NSRange adjustRange, NSRange editedRange, NSInteger delta );
NSRange NisusRange_AdjustedUnionForEdit( NSRange adjust, NSRange modRng, NSInteger delta );
NSRange NisusRange_AdjustForDeletion( NSRange adjust, NSRange removed );
NSRange NisusRange_AdjustForInsertion( NSRange adjust, NSRange inserted );


